from __future__ import annotations

from .data import Compound

__all__ = [
    "COMPARTMENT_MAP",
    "TYPE_MAP",
    "COMPARTMENT_SUFFIXES",
    "MANUAL_ADDITIONS",
]

COMPARTMENT_MAP = {
    "CYTOSOL": "CYTOSOL",
    "IN": "CYTOSOL",
    "UNKNOWN-SPACE": "CYTOSOL",
    "SIDE-1": "CYTOSOL",
    "SIDE-2": "EXTRACELLULAR",
    "EXTRACELLULAR": "EXTRACELLULAR",
    "CHL-THY-MEM": "PERIPLASM",
    "CHLOR-STR": "CYTOSOL",
    "CHROM-STR": "CYTOSOL",
    "GOLGI-LUM": "CYTOSOL",
    "LYS-LUM": "CYTOSOL",
    "MIT-IM-SPC": "CYTOSOL",
    "MIT-IMEM": "PERIPLASM",
    "MIT-LUM": "CYTOSOL",
    "OUTER-MEM": "PERIPLASM",
    "PERI-BAC": "PERIPLASM",
    "PERI-BAC-GN": "PERIPLASM",
    "PERIPLASM": "PERIPLASM",
    "PEROX-LUM": "CYTOSOL",
    "PLASMA-MEM": "PERIPLASM",
    "PLAST-IMEM": "PERIPLASM",
    "PLASTID-STR": "PERIPLASM",
    "PM-ANIMAL": "PERIPLASM",
    "PM-BAC-ACT": "PERIPLASM",
    "PM-BAC-NEG": "PERIPLASM",
    "PM-BAC-POS": "PERIPLASM",
    "RGH-ER-LUM": "CYTOSOL",
    "PM-FUNGI": "PERIPLASM",
    "RGH-ER-MEM": "PERIPLASM",
    "THY-LUM-CYA": "CYTOSOL",
    "VAC-LUM": "CYTOSOL",
    "VAC-MEM": "PERIPLASM",
    "VESICLE": "PERIPLASM",
    "OUT": "PERIPLASM",
    "THY-MEM-CYA": "PERIPLASM",
    "VESICLE-MEM": "PERIPLASM",
}

TYPE_MAP = {
    "ETR-Quinones": "Ubiquinones",
    "ETR-Quinols": "Ubiquinols",
}

COMPARTMENT_SUFFIXES = {
    # Common to all
    "CYTOSOL": "c",
    "EXTRACELLULAR": "e",
    "PERIPLASM": "p",
    "MITOCHONDRIA": "m",
    "PEROXISOME": "x",
    "ER": "r",
    "VACUOLE": "v",
    "NUCLEUS": "n",
    "GOLGI": "g",
    "THYLAKOID": "u",
    "LYSOSOME": "l",
    "CHLOROPLAST": "h",
    "FLAGELLUM": "f",
    "EYESPOT": "s",
    "INTERMEMBRANE": "im",
    "CARBOXYSOME": "cx",
    "THYLAKOID-MEMBRANE": "um",
    "CYTOSOLIC-MEMBRANE": "cm",
    "INNER-MITOCHONDRIA": "i",
    "MITOCHONDRIA-INNER-MEMBRANE": "mm",
    "WILDTYPE": "w",
    "CYTOCHROME-COMPLEX": "y",
}

MANUAL_ADDITIONS = {
    "Acceptor_c": Compound(
        base_id="Acceptor",
        id="Acceptor_c",
        compartment="CYTOSOL",
        charge=0,
        formula={"Unknown": 1},
    ),
    "Donor-H2_c": Compound(
        base_id="Donor-H2",
        id="Donor-H2_c",
        compartment="CYTOSOL",
        charge=0,
        formula={"Unknown": 1, "H": 2},
    ),
    "Oxidized-ferredoxins_c": Compound(
        base_id="Oxidized-ferredoxins",
        id="Oxidized-ferredoxins_c",
        compartment="CYTOSOL",
        charge=1,
        formula={"Unknown": 1},
    ),
    "Reduced-ferredoxins_c": Compound(
        base_id="Reduced-ferredoxins",
        id="Reduced-ferredoxins_c",
        compartment="CYTOSOL",
        charge=0,
        formula={"Unknown": 1},
    ),
    "Red-NADPH-Hemoprotein-Reductases_c": Compound(
        base_id="Red-NADPH-Hemoprotein-Reductases",
        id="Red-NADPH-Hemoprotein-Reductases_c",
        compartment="CYTOSOL",
        charge=0,
        formula={"Unknown": 1, "H": 2},
    ),
    "Ox-NADPH-Hemoprotein-Reductases_c": Compound(
        base_id="Ox-NADPH-Hemoprotein-Reductases",
        id="Ox-NADPH-Hemoprotein-Reductases_c",
        compartment="CYTOSOL",
        charge=0,
        formula={"Unknown": 1},
    ),
    "Cytochromes-C-Oxidized_c": Compound(
        base_id="Cytochromes-C-Oxidized",
        id="Cytochromes-C-Oxidized_c",
        compartment="CYTOSOL",
        charge=1,
        formula={"Unknown": 1},
    ),
    "Cytochromes-C-Reduced_c": Compound(
        base_id="Cytochromes-C-Reduced",
        id="Cytochromes-C-Reduced_c",
        compartment="CYTOSOL",
        charge=0,
        formula={"Unknown": 1},
    ),
    "Oxidized-Plastocyanins_c": Compound(
        base_id="Oxidized-Plastocyanins",
        id="Oxidized-Plastocyanins_c",
        compartment="CYTOSOL",
        charge=1,
        formula={"Unknown": 1},
    ),
    "Plastocyanin-Reduced_c": Compound(
        base_id="Plastocyanin-Reduced",
        id="Plastocyanin-Reduced_c",
        compartment="CYTOSOL",
        charge=0,
        formula={"Unknown": 1},
    ),
    "ETF-Oxidized_c": Compound(
        base_id="ETF-Oxidized",
        id="ETF-Oxidized_c",
        compartment="CYTOSOL",
        charge=1,
        formula={"Unknown": 1},
    ),
    "ETF-Reduced_c": Compound(
        base_id="ETF-Reduced",
        id="ETF-Reduced_c",
        compartment="CYTOSOL",
        charge=2,
        formula={"Unknown": 1, "H": 3},
    ),
    "Ox-Glutaredoxins_c": Compound(
        base_id="Ox-Glutaredoxins",
        id="Ox-Glutaredoxins_c",
        compartment="CYTOSOL",
        charge=1,
        formula={"Unknown": 1},
    ),
    "Red-Glutaredoxins_c": Compound(
        base_id="Red-Glutaredoxins",
        id="Red-Glutaredoxins_c",
        compartment="CYTOSOL",
        charge=0,
        formula={"Unknown": 1},
    ),
    "Ox-Thioredoxin_c": Compound(
        base_id="Ox-Thioredoxin",
        id="Ox-Thioredoxin_c",
        compartment="CYTOSOL",
        charge=0,
        formula={"Unknown": 1},
    ),
    "Red-Thioredoxin_c": Compound(
        base_id="Red-Thioredoxin",
        id="Red-Thioredoxin_c",
        compartment="CYTOSOL",
        charge=0,
        formula={"Unknown": 1, "H": 2},
    ),
    "Ox-FMN-Flavoproteins_c": Compound(
        base_id="Ox-FMN-Flavoproteins",
        id="Ox-FMN-Flavoproteins_c",
        compartment="CYTOSOL",
        charge=1,
        formula={"Unknown": 1},
    ),
    "Red-FMNH2-Flavoproteins_c": Compound(
        base_id="Red-FMNH2-Flavoproteins",
        id="Red-FMNH2-Flavoproteins_c",
        compartment="CYTOSOL",
        charge=0,
        formula={"Unknown": 1, "H": 2},
    ),
    "Ox-FAD-Flavoproteins_c": Compound(
        base_id="Ox-FAD-Flavoproteins",
        id="Ox-FAD-Flavoproteins_c",
        compartment="CYTOSOL",
        charge=1,
        formula={"Unknown": 1},
    ),
    "Red-FADH2-Flavoproteins_c": Compound(
        base_id="Red-FADH2-Flavoproteins",
        id="Red-FADH2-Flavoproteins_c",
        compartment="CYTOSOL",
        charge=0,
        formula={"Unknown": 1, "H": 2},
    ),
    "Light_c": Compound(
        base_id="Light",
        id="Light_c",
        compartment="CYTOSOL",
        charge=0,
        formula={"Unknown": 0},
    ),
    "Oxidized-flavodoxins_c": Compound(
        base_id="Oxidized-flavodoxins",
        id="Oxidized-flavodoxins_c",
        compartment="CYTOSOL",
        charge=0,
        formula={"Unknown": 1},
    ),
    "Reduced-flavodoxins_c": Compound(
        base_id="Reduced-flavodoxins",
        id="Reduced-flavodoxins_c",
        compartment="CYTOSOL",
        charge=0,
        formula={"Unknown": 1, "H": 2},
    ),
}



from __future__ import annotations

import logging
from functools import partial
from typing import Any, Callable, Dict, Iterable

from ..data import Reaction
from .shared import (
    MALFORMED_LINE_STARTS,
    _add_database_link,
    _add_type,
    _rename,
    _set_gibbs0,
    _set_name,
)

__all__ = ["parse_reactions"]

logger = logging.getLogger(__name__)
Action = Callable[[Any, Any, Any], None]
SubAction = Dict[str, Callable[[Any, Any, Any, Any], None]]


def _set_ec_number(reactions: Dict[str, Reaction], id_: str, ec_number: str) -> None:
    reactions[id_].ec = ec_number


def _add_reaction_pathway(reactions: Dict[str, Reaction], id_: str, pathway: str) -> None:
    reactions[id_].pathways.add(pathway)


def _add_enzrxn(reactions: Dict[str, Reaction], id_: str, enzrxn: str) -> None:
    reactions[id_].enzrxns.add(enzrxn)


def _set_reaction_direction(reactions: Dict[str, Reaction], id_: str, direction: str) -> None:
    reactions[id_].direction = direction
    if direction == "REVERSIBLE":
        reactions[id_].reversible = True
    else:
        reactions[id_].reversible = False


def _add_reaction_location(reactions: Dict[str, Reaction], id_: str, location: str) -> None:
    location = location.replace("CCI-", "CCO-")
    if location.startswith("CCO-"):
        reactions[id_].locations.append(location)


def _set_substrate(
    reactions: Dict[str, Reaction],
    id_: str,
    substrate: str,
    type_map: Dict[str, str],
) -> None:
    substrate = _rename(type_map.get(substrate, substrate)) + "_c"
    reactions[id_].substrates[substrate] = -1
    reactions[id_].substrate_compartments[substrate] = "CCO-IN"


def _set_product(
    reactions: Dict[str, Reaction],
    id_: str,
    product: str,
    type_map: Dict[str, str],
) -> None:
    product = _rename(type_map.get(product, product)) + "_c"
    reactions[id_].products[product] = 1
    reactions[id_].product_compartments[product] = "CCO-IN"


def _set_substrate_coefficient(
    reactions: Dict[str, Reaction],
    id_: str,
    coefficient: str,
    substrate: str,
    type_map: Dict[str, str],
) -> None:
    try:
        reactions[id_].substrates[_rename(type_map.get(substrate, substrate)) + "_c"] = -float(coefficient)
    except ValueError:  # conversion failed
        pass


def _set_product_coefficient(
    reactions: Dict[str, Reaction],
    id_: str,
    coefficient: str,
    product: str,
    type_map: Dict[str, str],
) -> None:
    try:
        reactions[id_].products[_rename(type_map.get(product, product)) + "_c"] = float(coefficient)
    except ValueError:  # conversion failed
        pass


def _set_substrate_compartment(
    reactions: Dict[str, Reaction],
    id_: str,
    compartment: str,
    substrate: str,
    type_map: Dict[str, str],
) -> None:
    substrate = _rename(type_map.get(substrate, substrate)) + "_c"
    if compartment == "CCO-OUT":
        reactions[id_].substrate_compartments[substrate] = compartment
    elif compartment == "CCO-MIDDLE":
        reactions[id_].substrate_compartments[substrate] = "CCO-OUT"
    else:
        # reactions[id_].product_compartments[substrate] = "CCO-OUT"
        logger.info(f"Unknown compartment {compartment}")


def _set_product_compartment(
    reactions: Dict[str, Reaction],
    id_: str,
    compartment: str,
    product: str,
    type_map: Dict[str, str],
) -> None:
    product = _rename(type_map.get(product, product)) + "_c"
    if compartment == "CCO-OUT":
        reactions[id_].product_compartments[product] = compartment
    elif compartment == "CCO-MIDDLE":
        reactions[id_].product_compartments[product] = "CCO-OUT"
    else:
        # reactions[id_].product_compartments[product] = "CCO-OUT"
        logger.info(f"Unknown compartment {compartment}")


def parse_reactions(
    file: Iterable[str],
    type_map: Dict[str, str],
    actions: Dict[str, Action] | None = None,
    sub_actions: Dict[str, SubAction] | None = None,
    db_to_resource: dict[str, str] | None = None,
) -> Dict[str, Reaction]:
    if db_to_resource is None:
        db_to_resource = {
            "BIGG": "bigg.reaction",
            "METANETX-RXN": "metanetx.reaction",
            "RHEA": "rhea",
            "PIR": "pirsf",
            "UNIPROT": "uniprot",
            "SEED": "seed.reaction",
        }

    if actions is None:
        actions = {
            "TYPES": partial(_add_type, type_map=type_map),
            "COMMON-NAME": _set_name,
            "DBLINKS": partial(_add_database_link, db_to_resource=db_to_resource),
            "EC-NUMBER": _set_ec_number,
            "ENZYMATIC-REACTION": _add_enzrxn,
            "GIBBS-0": _set_gibbs0,
            "IN-PATHWAY": _add_reaction_pathway,
            "LEFT": partial(_set_substrate, type_map=type_map),
            "REACTION-DIRECTION": _set_reaction_direction,
            "RIGHT": partial(_set_product, type_map=type_map),
            "RXN-LOCATIONS": _add_reaction_location,
        }

    if sub_actions is None:
        sub_actions = {
            "^COMPARTMENT": {
                "LEFT": partial(_set_substrate_compartment, type_map=type_map),
                "RIGHT": lambda a, b, c, d: _set_product_compartment(a, b, c, d, type_map=type_map),
            },
            "^COEFFICIENT": {
                "LEFT": lambda a, b, c, d: _set_substrate_coefficient(a, b, c, d, type_map=type_map),
                "RIGHT": lambda a, b, c, d: _set_product_coefficient(a, b, c, d, type_map=type_map),
            },
        }

    id_ = ""
    reactions = {}
    last_identifier = ""
    last_content = ""
    for line in file:
        if any(line.startswith(i) for i in MALFORMED_LINE_STARTS):
            continue
        try:
            identifier, content = line.rstrip().split(" - ", maxsplit=1)
        except ValueError:
            logger.info(f"Malformed line in reactions.dat: {line}")
            continue

        if identifier == "UNIQUE-ID":
            id_ = content
            reactions[id_] = Reaction(
                id=id_,
                base_id=id_,
                database_links={
                    "biocyc": set([id_]),
                    "metacyc.reaction": set([id_]),
                },
            )
        elif identifier.startswith("^"):
            if (subaction := sub_actions.get(identifier, {}).get(last_identifier, None)) is not None:
                subaction(reactions, id_, content, last_content)
        else:
            if (action := actions.get(identifier, None)) is not None:
                action(reactions, id_, content)
                last_identifier = identifier
                last_content = content

    return reactions



from __future__ import annotations

import warnings
from pathlib import Path
from typing import Dict, List

from ..data import Result
from .compounds import parse_compounds
from .enzymes import parse_enzymes
from .monomer_annotation import parse_monomer_annotation
from .monomer_sequences import parse_sequences
from .protein_complexes import parse_complexes
from .reactions import parse_reactions

__all__ = [
    "parse_pgdb",
    "parse_compounds",
    "parse_enzymes",
    "parse_monomer_annotation",
    "parse_complexes",
    "parse_reactions",
    "parse_sequences",
]


def remove_top_comments(file: List[str]) -> List[str]:
    """Remove the metainformation from a pgdb file."""
    for i, line in enumerate(file):
        if line.startswith("UNIQUE-ID"):
            return file[i:]
    return file


def read_file(path: Path) -> List[str]:
    try:
        with open(path, encoding="ISO-8859-14") as f:
            return f.readlines()
    except FileNotFoundError:
        warnings.warn(f"Could not find file {path}", stacklevel=2)
        return []


def read_file_and_remove_comments(path: Path) -> List[str]:
    """Read the file and remove metainformation."""
    return remove_top_comments(read_file(path))


def parse_pgdb(path: Path, type_map: Dict[str, str] | None = None) -> Result:
    if type_map is None:
        type_map = {}

    return Result(
        compounds=parse_compounds(read_file_and_remove_comments(path / "compounds.dat"), type_map),
        reactions=parse_reactions(read_file_and_remove_comments(path / "reactions.dat"), type_map),
        monomer_annotation=parse_monomer_annotation(read_file_and_remove_comments(path / "genes.dat")),
        complexes=parse_complexes(read_file_and_remove_comments(path / "proteins.dat")),
        enzymes=parse_enzymes(read_file_and_remove_comments(path / "enzrxns.dat")),
        sequences=parse_sequences(iter(read_file(path / "protseq.fsa"))),
    )



from __future__ import annotations

import logging
from typing import Any, Callable, Dict, Iterable, Set

from .shared import MALFORMED_LINE_STARTS

logger = logging.getLogger(__name__)

__all__ = ["parse_complexes"]
Action = Callable[[Any, Any, Any], None]


def _add_component(complexes: Dict[str, Set[str]], complex_id: str, component: str) -> None:
    complexes[complex_id].add(component)


def parse_complexes(
    file: Iterable[str],
    actions: Dict[str, Action] | None = None,
) -> Dict[str, Set[str]]:

    if actions is None:
        actions = {
            "COMPONENTS": _add_component,
        }

    id_ = ""
    proteins: Dict[str, Set[str]] = {}
    complexes: Dict[str, Set[str]] = dict()
    for line in file:
        if any(line.startswith(i) for i in MALFORMED_LINE_STARTS):
            continue
        try:
            identifier, content = line.rstrip().split(" - ", maxsplit=1)
        except ValueError:
            logger.info(f"Malformed line in proteins.dat: {line}")
            continue

        if identifier == "UNIQUE-ID":
            id_ = content
            proteins[id_] = set()
        elif not identifier.startswith("^"):
            if (action := actions.get(identifier, None)) is not None:
                action(proteins, id_, content)

    for k, v in proteins.items():
        if bool(v):
            complexes[k] = v
    return complexes



from __future__ import annotations

import logging
from functools import partial
from typing import Any, Callable, Dict, Iterable

from ..data import Compound
from .shared import (
    MALFORMED_LINE_STARTS,
    _add_database_link,
    _add_type,
    _set_gibbs0,
    _set_name,
)

__all__ = ["parse_compounds"]

logger = logging.getLogger(__name__)
Action = Callable[[Dict[str, Any], str, str], None]


def _set_atom_charges(compounds: Dict[str, Compound], id_: str, content: str) -> None:
    try:
        compounds[id_].charge += int(content[1:-1].split()[-1])
    except ValueError:  # conversion failed
        pass


def _set_chemical_formula(compounds: Dict[str, Compound], id_: str, content: str) -> None:
    atom, count = content[1:-1].split(" ")
    try:
        compounds[id_].formula[atom] = int(count)
    except ValueError:  # conversion failed
        pass


def _set_smiles(compounds: Dict[str, Compound], id_: str, content: str) -> None:
    compounds[id_].smiles = content


def parse_compounds(
    file: Iterable[str],
    type_map: dict[str, str],
    actions: Dict[str, Action] | None = None,
    db_to_resource: dict[str, str] | None = None,
) -> Dict[str, Compound]:
    if db_to_resource is None:
        db_to_resource = {
            "BIGG": "bigg.metabolite",
            "BRENDA-COMPOUND": "brenda",
            "CAS": "cas",
            "CHEBI": "CHEBI",
            "CHEMSPIDER": "CHEMSPIDER",
            "DRUGBANK": "drugbank",
            # "ECOCYC": None,
            "HMDB": "HMDB",
            "KEGG": "kegg.compound",
            "KEGG-GLYCAN": "kegg.glycan",
            "KNAPSACK": "knapsack",
            # "LIGAND-CPD": None,
            # "LIPID_MAPS": None,
            # "MEDIADB": None,
            "METABOLIGHTS": "metabolights",
            "METANETX": "metanetx.chemical",
            # "NCI": None,
            "PUBCHEM": "pubchem.compound",
            "PUBCHEM-SID": "pubchem.substance",
            "REACTOME-CPD": "reactome",
            # "REFMET": None,
            "SEED": "seed.compound",
            "UM-BBD-CPD": "umbbd.compound",
        }
    if actions is None:
        actions = {
            "TYPES": partial(_add_type, type_map=type_map),
            "COMMON-NAME": _set_name,
            "ATOM-CHARGES": _set_atom_charges,
            "CHEMICAL-FORMULA": _set_chemical_formula,
            "DBLINKS": partial(_add_database_link, db_to_resource=db_to_resource),
            "GIBBS-0": _set_gibbs0,
            "SMILES": _set_smiles,
        }

    compounds: Dict[str, Compound] = {}
    id_ = ""
    for line in file:
        if any(line.startswith(i) for i in MALFORMED_LINE_STARTS):
            continue
        try:
            identifier, content = line.rstrip().split(" - ", maxsplit=1)
        except ValueError:
            logger.info(f"Malformed line in compoudns.dat: {line}")
            continue

        if identifier == "UNIQUE-ID":
            base_id = content
            id_ = content + "_c"
            compounds[id_] = Compound(
                id=id_,
                base_id=base_id,
                compartment="CYTOSOL",
                database_links={
                    "biocyc": set([base_id]),
                    "metacyc.compound": set([base_id]),
                },
            )
        else:
            if (action := actions.get(identifier, None)) is not None:
                action(compounds, id_, content)
    return compounds



from __future__ import annotations

from typing import Dict, List, Set

from ..data import Enzyme, Reaction


def _get_monomers(
    cplx_or_monomer: str,
    complexes: Dict[str, Set[str]],
) -> Set[str]:
    result: Set[str] = set()
    if (components := complexes.get(cplx_or_monomer)) is not None:
        for component in components:
            if component in complexes:
                result.update(_get_monomers(component, complexes))
            else:
                result.add(component)
    else:  # doesn't necessarily need to be in monomers nor sequences
        result.add(cplx_or_monomer)
    return result


def get_gpr_associations(
    reactions: Dict[str, Reaction],
    enzymes: Dict[str, Enzyme],
    complexes: Dict[str, Set[str]],
) -> Dict[str, List[Set[str]]]:
    res: Dict[str, List[Set[str]]] = {}
    for name, reaction in reactions.items():
        for enzyme in reaction.enzrxns:
            if (enzrxn := enzymes.get(enzyme, None)) is not None:
                if (cplx := enzrxn.cplx_or_monomer) is not None:
                    res.setdefault(name, []).append(_get_monomers(cplx, complexes))
    return res



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['cycparser', 'cycparser.parsing', 'cycparser.repairing']

package_data = \
{'': ['*']}

install_requires = \
['gitchangelog>=3.0.4,<4.0.0']

setup_kwargs = {
    'name': 'cycparser',
    'version': '1.2.23',
    'description': '',
    'long_description': 'None',
    'author': 'Marvin van Aalst',
    'author_email': 'marvin.vanaalst@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import re
from typing import Dict, Iterator

__all__ = ["parse_sequences"]


def parse_sequences(file: Iterator[str]) -> Dict[str, str]:
    RE_PAT = re.compile(r"^>gnl\|.*?\|")
    sequences: Dict[str, str] = {}
    while True:
        try:
            id_ = re.sub(RE_PAT, "", next(file)).split(" ", maxsplit=1)[0]
            sequence = next(file).strip()
            sequences[id_] = sequence
        except StopIteration:
            break
    return sequences
